#!/usr/bin/perl

my $script = "./autoClient.pl";

my $PWD = $ENV{PWD};
use FileHandle;
use IPC::Open2;

#-----------------------------------------------------------------------
# getnum and is_numeric
# borrowed from: http://www.unix.org.ua/orelly/perl/cookbook/ch02_02.htm
#-----------------------------------------------------------------------

#-----------------------------------------------------------------------
# @brief obtain the numeric value of the string, decimal or integer
# @param str string to extract value from
# @return numeric value or undefined if no value extractable
#-----------------------------------------------------------------------
sub getnum {
    use POSIX qw(strtod);
    my $str = shift;
    $str =~ s/^\s+//;
    $str =~ s/\s+$//;
    $! = 0;
    my($num, $unparsed) = strtod($str);
    if (($str eq '') || ($unparsed != 0) || $!) {
        return;
    } else {
        return $num;
    } 
}

#-----------------------------------------------------------------------
# @brief checks if the string is numeric
# @param str string to check
# @return true if numeric, false otherwise
#-----------------------------------------------------------------------
sub is_numeric { defined scalar &getnum } 

sub getPositiveInt{
    my $str = shift;
    if($str =~ /^\s*\d+\s*$/){
	return getnum($str);
    }
    else{
	return;
    }
}

my @hostnames = qw(chess drlucky girltalk go 
		  othello risk roborally settlers stratego boggle devilbunny clue); #mahjongg arkhamhorror diplomacy

my @clientIDs = qw(1 2 3 4 5 6 7 8 9 10);
my @jboss_servers = qw(diplomacy settlers);
my $jndi_server = "mahjongg";
my $verbose = 1;

sub runTest{
    my ($design, $design_count, $serverCount, $clientCount, $transactionCount) = @_;
    
    print "Test($design/$design_count)\n";
    #disable buffering
    $| = 1;

    select STDOUT; $| = 1;

    my @children = ();

    #usage: java AutoClient <JNDI server> <auto-cli filename> <auto-cli ID> <numIterations>
    for (my $i = 0; $i < $clientCount; $i++){
	my $pid;
	if($pid = fork()){
	    unshift(@children, $pid);
	}else{ #child
	    
	    my $clientID = $clientIDs[$i];
	    my $hostname = $hostnames[$i];
	    
	    if($verbose){
		print "--Running client(${i}): ./runAutoClient $jndi_server $clientCount"."cli_$hostname".".txt ".($i+1)." ".$transactionCount . "\n";
	    }
	    
	    if(!open(AUTOCLIENT, "rsh -f $hostname \" cd ${PWD}; ./runAutoClient $jndi_server $clientCount"."cli_$hostname".".txt ".($i+1). " " . $transactionCount . " \" 2>&1 |")){
		print "Unable to start autoClient $i on $hostname\n";
		return -1;
	    }
	
            while(defined(my $line = <AUTOCLIENT>)){
              print "--AUTOCLIENT($i)> " .  $line if(verbose);
            }
            close(AUTOCLIENT);
            exit(0);
          }
      }

    while(scalar(@children)){
	waitpid(shift(@children),0);
    }

    return 0;
}

#-----------------------------------------------------------------------
# @brief puts it all together, reading input, etc
#-----------------------------------------------------------------------
sub main {

    #cmdline args = defaults:
    my $inputFile = "experimentList";
    my $outDir = "temp";
    my $do_nothing = 0;

    #parse args
    for(my $i = 0; $i < scalar(@ARGV); $i++){
	if($ARGV[$i] eq ""){
	    next;
	}elsif($ARGV[$i] eq "-n"){
	    $do_nothing = 1;
	}else{
	    print "Unrecognized option: " . $ARGV[$i] . "\n";
	    printUsage();
	}
    }
    
    my @serverCounts = ();
    my @clientCounts = ();
    my @transactionCounts = ();

    open(PARAMS, "< $inputFile") || die "Unable to open input file: $!\n";

    while(defined(my $line = <PARAMS>)){

	chomp($line);	

	#remove any comments
	$line =~ s/\#.*$//;
	
	#remove any whitespace
	$line =~ s/^\s+//;
	$line =~ s/\s+$//;

	if( $line =~ /(\w+)\((.+)\)/ ){
	    my ($param, $vals) = ($1,$2);
	    
	    my @val_list = ();
	    my @splitVals = split(/,/, $vals);
	    foreach my $val (@splitVals) {
		unshift(@val_list, getnum($val));
	    }

	    #// should look sort of like:
	    #clients(1,4,7,10) #per tudor's email
	    #servers(2) #primary and backup
            #invocations(10000) #hardcoded --don't change

	    if($param eq "clients"){
		unshift(@clientCounts, @val_list);
	    }elsif($param eq "servers"){
		unshift(@serverCounts, @val_list);
	    }elsif($param eq "invocations"){
		unshift(@transactionCounts, @val_list);
	    }else{
		print STDERR "Unkown parameter $param\n";
		return -1;
	    }
	}
    }
    
    close(PARAMS);
    
    my $design_count = 0;
    
    {#local scope variables
	 foreach my $serverCount(@serverCounts){
	     foreach my $clientCount(@clientCounts){
               foreach my $transactionCount(@transactionCounts){
                 $design_count++;
               }
             }
           }
       }
    
    if($design_count == 0){
	print "ERROR: No tests to iterate on.  Perhaps one of the options was left out\n";
	return -1;
    }

    print "Iterating on $design_count tests\n";
    
    my $design = 0;

    {#local scope variables
	 foreach my $serverCount(@serverCounts){
	     foreach my $clientCount(@clientCounts){
               foreach my $transactionCount(@transactionCounts){
                 $design++;
                 
                 if($do_nothing){
                   print "runTest(test = $design/$design_count, serverCount = $serverCount, clientCount = $clientCount, transactionCount = $transactionCount);\n";
                 }else{
                   $done = -1;
                   while($done == -1) {
                     $done = runTest($design, $design_count, $serverCount, $clientCount, $transactionCount);
                     if($done < 0){
                       print "Test $design/$design_count FAILED, retrying...\n";
                     }
                   }
                 }
               }
             }
           }
       }

    print "Tests complete\n";

    return 0;
}

#-----------------------------------------------------------------------
# @brief prints the usage and exits -1
#-----------------------------------------------------------------------
sub printUsage(){
    
    print <<EOF;
    
$scriptInfo
    
    $script 
       USAGE
       Don't do stupid stuff
       deletes the whole temp directory and structure, runs tests, pulls results from that directory
       also deletes the reports directory and structure then replaces it

EOF

  exit(-1);
}

if(main()){
    printf("\n\nError: Script Failed.\n");
    exit(-1);
}
else{
    exit(0);
}

1;
