
package ejbay.replication;

import javax.naming.InitialContext;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.Binding;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import java.io.*;
import java.util.Properties;
import java.lang.*;
import java.util.*;
//import java.awt.*;
//import java.awt.event.*;
import ejbay.Probe.*;
import ejbay.util.*;

import java.lang.reflect.Array;

public class FaultInjector extends Thread {
    
    /**
     * Constructor initializes necessary variables
     *
     * @param hbs The replication managers static arrayList
     *            of heartbeat threads
     * @param jndi The global JNDI server's location
     * @param t The replication manager's timer so that
     *          all timing is synchronized.
     */
    public FaultInjector(ArrayList hbs, String jndi, Time t) {
        hbThreads = hbs;
        time = t;
        probe1 = new Probe("fault_injection_time.txt");

        // Conect to the global JNDI server that should have been passed in
        try {
            globalJndi = getInitialContext(jndi, GLOBAL_JNDI_PORT);
            
        }catch(javax.naming.NamingException ne){
            System.out.println("Error: Could not connect to the global JNDI server");
            System.out.println("       make sure the server is started and the\n" +
                               "       proper address was given");
            loop = false;
        }
    }
    
    /**
     * Main loop will run a command parser along with various other
     * things that are needed for fault injection.
     */
    public void run() {
        
        BufferedReader cmdLine=new BufferedReader(new InputStreamReader(System.in));
        
        int cmd = -1;
        
        System.out.println("Welcome!");
        printHelp();
        System.out.println("Please enter your command");
        
        while(cmd != CMD_QUIT && loop) {
            try {
                String cmdString;
                
                System.out.print("> ");
                // Read input from command line
                cmdString = cmdLine.readLine();
                cmdString = cmdString.trim();
                if(cmdString.equals(""))
                    continue;
                // Parse input for the request
                cmd = parseRequest(cmdString);
                // Handle that request
                handleRequest(cmd);
            } catch (java.io.IOException ioe) {
                ioe.printStackTrace();
                System.out.println("Error processing your input... terminating servers");
                return;
            }
        }
        cancelKillTimer();
        probe1.output();
    }
    
    /**
     * Stops this tread from executing and cleans up
     * any running things
     */
    public void destroy() {
        loop = false;
        cancelKillTimer();
    }
    
    /** 
     * Prompts the standard input using a provided list of questions and
     * returns a list of the users answers for processing by the application
     *
     * @param questions A list of the questions to be asked in a string array
     * @return A list of the answers to the given questions as a string array
     */
    public String[] AskUser(String[] questions) {
        String[] answers = new String[questions.length];
        BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
        
        try {
            String inputString;
            for (int i = 0; i < questions.length; i++) {
                System.out.print(questions[i]);
                inputString = input.readLine();
                if (inputString.length() == 0)
                    inputString = " ";
                answers[i] = inputString;
            }
        } catch (IOException ioe) {
            System.out.println("IO error trying to read your input!");
            System.exit(1);
        }
        return answers;
    }
    
    /** 
     * Get a new initial context that sets up the client to find
     * the JNDI server.
     *
     * @param jndiServer The URL of the JNDI server to connect to
     * @param port The default port for this connection
     * @return A handle to the context that JNDI should use
     */
    public Context getInitialContext(String jndiServer, int port) throws javax.naming.NamingException {
        Properties env = new Properties();
        
        if(jndiServer.indexOf(":") == -1)
            jndiServer = jndiServer + ":" + Integer.toString(port);
        
        env.setProperty(Context.INITIAL_CONTEXT_FACTORY,
                "org.jnp.interfaces.NamingContextFactory");
        env.setProperty(Context.URL_PKG_PREFIXES,
                "org.jboss.naming:org.jnp.interfaces");
        env.setProperty(Context.PROVIDER_URL, jndiServer);
        
        return new InitialContext(env);
    }
    
    /** 
     * Take in a command line request and find what request is being made
     *
     * @param req The input request
     * @return The integer command value of the request
     */
    public int parseRequest(String req) {
        // Handle basic shell commands
        if(req.equalsIgnoreCase("help") || req.equalsIgnoreCase("h"))
            return CMD_HELP;
        if(req.equalsIgnoreCase("quit")
        || req.equalsIgnoreCase("exit")
        || req.equalsIgnoreCase("q"))
            return CMD_QUIT;
        
        // Handle application commands
        if(req.equalsIgnoreCase("list") || req.equalsIgnoreCase("ls"))
            return CMD_LIST;
        if(req.equalsIgnoreCase("listjndi") || req.equalsIgnoreCase("lsjndi"))
            return CMD_LIST_JNDI;
        if(req.equalsIgnoreCase("kill") || req.equalsIgnoreCase("k"))
            return CMD_KILL;
        if(req.equalsIgnoreCase("sched_kill") || req.equalsIgnoreCase("sk"))
            return CMD_START_PERIODIC;
        if(req.equalsIgnoreCase("cancel_kill") || req.equalsIgnoreCase("ck"))
            return CMD_STOP_PERIODIC;

        return CMD_UNKNOWN;
    }
    
    /** 
     * Performs the necessary action and function call for a given command
     * message.
     *
     * @param cmd Command Message ID number to process
     */
    public void handleRequest(int cmd) throws java.rmi.RemoteException {
        // Handle shell commands
        switch(cmd) {
            case CMD_QUIT:         // Quit
                return;
            case CMD_UNKNOWN:      // Unknown
                System.out.println("Error processing request!  Check your input and try again");
                return;
            case CMD_HELP:         // Help
                printHelp();
                return;
        }
        
        // Make sure there's a recent server list before any command
        updateServerList();
        // Process Account and Auction Commands
        switch(cmd) {
            case CMD_LIST:          // List current replicas
                listReplicas();
                break;
            case CMD_LIST_JNDI:     // List JNDI context
                listJndi();
                break;
            case CMD_KILL:          // Kill Server
                killReplica();
                break;
            case CMD_START_PERIODIC:
                scheduleKillTimer();
                break;
            case CMD_STOP_PERIODIC:
                cancelKillTimer();
                break;
            default:                // Hopefully we never reach here
                System.out.println("Error processing request");
                break;
        }
    }
    
    /**
     * Prints the help message which contains a list of all possible command that this
     * shell will process.
     */
    public void printHelp() {
        System.out.println("List of available commands:");
        System.out.println("");
        System.out.println("kill               - kill a server");
        System.out.println("sched_kill(sk)     - set up a schedule to kill servers");
        System.out.println("cancel_kill(ck)    - stop periodic killing of servers");
        System.out.println("list               - list current replicas");
        System.out.println("listjndi           - list contents of the jndi server");
        System.out.println("help               - display this message");
        System.out.println("quit               - exit this prompt");
        System.out.println("****************************************************\n");
    }  
    
    /**
     * Update the list of current servers to match what the
     * replication manager currently has running. Do this
     * by reading the static hearbeat threads list.
     */
    public void updateServerList() {
        ServList nList = new ServList();
        ArrayList listCopy = new ArrayList(hbThreads); //Copy list
        
        // Use the heartbeat threads to list running servers
        for(int i = 0; i < listCopy.size(); i++) {
            HeartBeat temp = (HeartBeat)listCopy.get(i);
            
            if(temp != null) {
                RepCommonData replica = temp.getReplica();
                nList.addServer(replica, i);
            }
        }
        
        serverList = nList;
    }
    
    /** 
     * List all of the currently running replicas by going
     * through the locally stored server list
     */
    public void listReplicas() {
        int servers = serverList.numServers();
        
        for(int i=0; i < servers; i++) {
            System.out.print("Server " + i + ": ");
            if(i < 10)
                System.out.print(" ");
            System.out.println(serverList.findServer(i));
        }
    }
    
    /**
     * List all of the bindings stored in the global JNDI
     * so that we can see exactly what's going on.
     */
    public void listJndi() {
        try {
            NamingEnumeration enum = globalJndi.listBindings("");
            
            // Go through all bindings and print
            while (enum.hasMore()) {
                Binding binding = (Binding) enum.next();
                System.out.println(binding.getName() + " --> " + binding.getObject());
            }
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }
    
    /**
     * Prompt the user as to which replica they would like to kill
     * and then proceed to kill off the selected replica.
     */
    public void killReplica() {
        listReplicas();
        
        String[] q  = new String[1];
        q[0] = "Which replica would you like to kill(0 -> " + (serverList.numServers()-1) + "): ";
        String[] a = AskUser(q);
        
        try {
            int num = (new Integer(a[0])).intValue();
            RepCommonData repData = serverList.findServer(num);
            
            if(repData != null)
                killReplica(repData);
            else {
                System.out.println("That is not a valid server number");
                return;
            }
        } catch(NumberFormatException nfe) {
            System.out.println("Error: " + a[0] + " is not a number");
            return;
        }
    }
    
    /**
     * Kill the replica that matches the given replica data
     * object.
     *
     * @param repData The replica to kill
     */
    public void killReplica(RepCommonData repData) {
        try {
            String cmd = KILL_CMD + " " + repData.getServerName();
            //System.out.println("Trying: "+ cmd);
            Process p = Runtime.getRuntime().exec(cmd);            
            p.waitFor();

            probe1.record(String.valueOf(time.getElapsedTime()));
        }catch(Exception e) {
            System.out.println("Could not kill server " + repData.getServerName());
            return;
        }
        
        System.out.println("Killed replica " + repData.getServerName());
    }
    
    /**
     * Prompt the user for the various parameters in a kill timer
     * and schedule the server to be killed periodically based on
     * the users input.
     */
    public void scheduleKillTimer() {
        if (periodicKillTimer != null) {
            System.out.println("Periodic Timer is already running. Please stop it first.");
            return;
        }
        
        // Ask user what they want to do
        String[] q  = new String[3];
        q[0] = "How often would you like a server to be killed (seconds)? ";
        q[1] = "If you would like a random deviation from this period enter it,\notherwise use zero (milliseconds)? ";
        q[2] = "Would you like to kill the primary only (y/n)? ";
        String[] a = AskUser(q);
        
        try {
            int period = (new Integer(a[0])).intValue()*1000;
            int var = (new Integer(a[1])).intValue();
            boolean prim;
            if(a[2].equalsIgnoreCase("y"))
                prim = true;
            else if(a[2].equalsIgnoreCase("n"))
                prim = false;
            else
            {
                System.out.println("Error: please only enter y or n for the kill primary question");
                return;
            }
            // Create and schedule the timer task
            periodicKillTimer = new Timer();
            periodicKillTimer.schedule(new KillReplica_Timer(var, prim), KILL_DELAY*1000, period);

            if(prim)
                System.out.println("The primary server will be killed every " + period/1000 +
                                   " seconds, with variance " + var + "ms");
            else
                System.out.println("A random server will be killed every " + period/1000 +
                                   " seconds, with variance " + var + "ms");

            System.out.println("Killing will begin after a " + KILL_DELAY + " second delay");
            // Can execute twice if action was delayed
            // First execution occurs after 1 second delay.
            // periodicKilltimer.schedule(classofsomesort, 1000, period) // Subject to drift
            
        } catch(NumberFormatException nfe) {
            System.out.println("Error: " + a[0] + " and/or " + a[1] + " are not numbers");
            return;
        }
    }
    
    /**
     * If there is a currently running kill timer, then stop
     * that timer and return.
     */
    private void cancelKillTimer() {
        if(periodicKillTimer != null)  //Timer exists
        {
            periodicKillTimer.cancel();
            periodicKillTimer = null;
            System.out.println("Periodic killing stopped");
        }
    }

    /**
     * A class for the timerTask to run. This will store the
     * various kill parameters and kill a server every time
     * it is invoked.
     */
    class KillReplica_Timer extends TimerTask {
        /**
         * Setup this kill timer's parameters
         *
         * @param var The random deviation in killing
         * @param killPrim Whether or not only the primary
         *                 should be killed
         */
        public KillReplica_Timer(int var, boolean killPrim)
        {
            variance = var;
            killPrimary = killPrim;
        }

        /** 
         * Main method will perform the necessary killing
         */
        public void run() {
            RepCommonData repData;

            // Get a good server list
            updateServerList();

            // Find the replica to kill
            if(killPrimary)
                repData = serverList.getPrimary();
            else
                repData = serverList.findUsedServer();

            try {
                if (variance != 0) {  // Wait for a random bit of time                   
                    //System.out.println("variance = " + variance);
                    Random generator = new Random(System.currentTimeMillis());
                    int temp = generator.nextInt(variance);
                    //System.out.println("Sleeping for " + temp/1000.0 + " seconds");
                    sleep(temp);
                }
                
                // Kill the replica
                if(repData != null)
                    killReplica(repData);
                else {
                    System.out.println("No servers to kill...stopping auto killer");
                    cancelKillTimer();
                    return;
                }
            } catch (InterruptedException err) {}
        }
        
        private int variance;
        private boolean killPrimary;
    }    
    
    // Global connection variables
    static Context globalJndi;
    public ServList serverList;
    public static ArrayList hbThreads;
    public boolean loop = true;
    private Timer periodicKillTimer = null;
    private Time time;
    private Probe probe1;

    // Define constants for the different commands that are processed
    private static final int CMD_UNKNOWN        = -1;
    private static final int CMD_QUIT           = 0;
    private static final int CMD_HELP           = 1;
    private static final int CMD_LIST           = 2;
    private static final int CMD_KILL           = 3;
    private static final int CMD_LIST_JNDI      = 5;    
    private static final int CMD_START_PERIODIC = 6;
    private static final int CMD_STOP_PERIODIC  = 7;
    
    private static final int KILL_DELAY         = 6;

    private static String CWD                   = System.getProperty("user.dir", "");
    private static String KILL_CMD              = CWD + "/killAllGamesJava";
    private static final int LOCAL_JNDI_PORT    = 6729;
    private static final int GLOBAL_JNDI_PORT   = 59080;
}

