package ejbay.replication;

import ejbay.account.AccountHomeRemote;
import ejbay.account.AccountRemote;
import ejbay.auction.*;
import ejbay.Probe.*;
import ejbay.util.*;

import javax.naming.InitialContext;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import java.io.*;
import java.util.Properties;
import java.lang.*;
import java.util.*;


public class HeartBeat extends Thread {

    /**
     * Constructor sets all of this heartBeats global data so it knows what server
     * to monitor
     *
     * @param repData A data object representing the server to monitor
     * @param glblJndi A handle to the global JNDI
     * @param str A string representing this server's name in the global JNDI
     * @param dbg A parameter that specifies whether to print the output from this
     *            server to the screen
     * @param t The timer variable to be used for recording timing data
     */
    public HeartBeat(RepCommonData repData, Context glblJndi, String str, boolean dbg, Time t) 
    {
        myReplica = repData;
        globalJndi = glblJndi;
        jndiName = str;
        debug = dbg;
        time = t;
        probe1 = new Probe("hb_fault_detected.txt");
    }

    /**
     * Runs the main method for heartbeating which is responsible for monitoring
     * a server as it boots, updating the global JNDI, and heartbeating this
     * server until it finishes.
     */
    public void run() 
    {        
        boolean notStarted = true;
        Process p = myReplica.getServerProc();
        InputStream inStream = p.getInputStream();
        InputStreamReader reader = new InputStreamReader(inStream);
        BufferedReader input = new BufferedReader(reader);

        /* If debug mode, then just grab input and send it to the screen */
        if(debug)
        {
            while(loop)
            {
                try
                {
                    String temp = input.readLine();
                
		    // Print out all messages b/c this is a debug server
                    if(temp != null)
                        System.out.println(temp);                        
                    else
		    {
			cod = COD_UNKNOWN;
                        return;
		    }

                } catch(java.io.IOException ioe) {
		    cod = COD_EXCEPTION;
                    return;                
                }
            }
        }

        // Wait for the jboss server to start
        while(notStarted && loop)
        {
            try
            {
                String temp = input.readLine();
                
                if(temp != null)
                {		   
		    // If JBoss reports the server has started then...
                    if(temp.indexOf(JBOSS_STARTED) != -1)
                    {
                        System.out.println("Server: " + myReplica.getServerName() + " booted in " +
                                           temp.substring(temp.indexOf("Started in")+11));
                        myReplica.setBooted();
                        notStarted = false;
                    }
		    // If Jboss reports a java.net.BindException then...
                    if(temp.indexOf(SERVER_EXCEPTION) != -1)
                    {
                        System.out.println("Failed to start server " + myReplica.getServerName() + 
                                           " because a server is already running on this machine");                        
                        loop = false;
			cod = COD_EXCEPTION;
                    }                    
                }
                else
                {
		    System.out.println("Error: Server " + myReplica.getServerName() + " not started properly");
                    loop = false;
		    cod = COD_UNKNOWN;		   
                }
            } catch(java.io.IOException ioe) {
		cod = COD_IMPRSTART;
                return;                
            }
        } 

	// The server is now booted, so pause for a second and then try to connect
	// to make sure it's ready
        try
        {
            Thread.sleep(1000);
        }catch(InterruptedException ie){}
        
	// Try to open a connection to this server
        if(!startHeartBeatConn())
        {
            cod = COD_IMPRSTART;
            System.out.println("Error: Server " + myReplica.getServerName() + " did not start correctly");
            return;
        }
        
        // Add my replica to the global JNDI
        try
        {           
            globalJndi.rebind(jndiName, myReplica.getServerName()+":"+Integer.toString(LOCAL_JNDI_PORT));
        } catch(Exception e) {
            System.out.println("Error: Server " + myReplica.getServerName() +
                               " could not be added to the global JNDI");
        }

        // Heartbeat until we're destroyed or the server dies
        while(loop)
        {         
            try {
                sleep(HEARTBEAT_PERIOD);
            } catch (InterruptedException e) {}

            /* Check to see if the server is still running */
	    if(!serverAlive() && loop)
	    {
                probe1.record(String.valueOf(time.getElapsedTime()));
                probe1.output();
		cod = COD_SCRASH;
		System.out.println("[HB] Server: " + myReplica.getServerName() + " server crashed, launching a new replica...");
		return;
	    }
	    else if(!processAlive() && loop)  /* Make sure the process is still running */
	    {
                probe1.record(String.valueOf(time.getElapsedTime()));
                probe1.output();
		cod = COD_PCRASH;
		System.out.println("[HB] Server: " + myReplica.getServerName() + " process crashed, launching a new replica...");
		return;
	    }
        }
    }

    /* Try to detect whether a server is alive by remmotely executing the hostname function and
     * making sure that it matches the known hostname value
     *
     * @return True if this thread's replica's server is still alive. False otherwise
     */
    public boolean serverAlive()
    {
	String hostResponse;
	String serverName = myReplica.getServerName();

	if(serverName == null || serverName.equals(""))
	    return false;

	// Run the 'hostname' command on the remote server and examine it's
	// output to see if it matches what we know to be the hostname
	try
	{
	    Process p = Runtime.getRuntime().exec("/usr/local/bin/rsh " + serverName + " hostname");
	    p.waitFor();
	    InputStream inStream = p.getInputStream();
	    InputStreamReader reader = new InputStreamReader(inStream);
	    BufferedReader input = new BufferedReader(reader);

	    hostResponse = input.readLine();

	    // Check the two strings, but only compare up to the first period in the
	    // server's DNS name
            int hostLen = hostResponse.length();
            int servLen = serverName.length();
            if(servLen > hostLen)
                serverName = serverName.substring(0,hostLen);
            else if(hostLen != servLen)
                return false;

	    if(!hostResponse.equalsIgnoreCase(serverName))
		return false;

	}catch(Exception e){
	    return false;
	}
	
	return true;
    }

    /* Try to detect whether the bean processes are still running on the server
     * do this by invoking the isAlive() method on each bean.
     *
     * @return True if this thread's replica is still alive. False otherwise
     */
    public boolean processAlive()
    {
	boolean alive = true;

	try
	{
	    // Call each bean's isAlive() method
	    acctBean.isAlive();
            aucBean.isAlive();
	    
	} catch(Exception e) {
	    alive = false;
	} 

	return alive;
    }
    
    /**
     * This function will destroy this running thread by setting a loop variable to
     * false and causing the main loop above to exit.
     */
    public void destroy()
    {
        loop = false;
	this.interrupt();
    }

    /**
     * Returns the recorded cause of death for this thread. This value should
     * have been set at exit from the main loop by this thread.
     *
     * @return An integer value corresponding to a specific cause of death
     */
    public int getCauseOfDeath()
    {
	return cod;
    }

    /**
    * Returns a copy of the replica object representing this heart beat thread's
    * server
    *
    * @return The replica representing this heartbeat thread
    */
    public RepCommonData getReplica()
    {
        return myReplica;
    }

    /**
     * Returns whether or not this thread's replica is currently booted
     *
     * @return True if this replica is booted, false otherwise
     */
    public boolean isBooted()
    {
        return myReplica.isBooted();
    }

    /* Initialize global objects that reference each server bean. Once these references
     * are established, the heartbeat thread can just call a .isAlive() mehtod on them.
     *
     * @return True if a connection was established with this replica, false otherwise.
     */
    public boolean startHeartBeatConn()
    {
        boolean alive = true;

	try
	{
	    // Connect to the Account Bean
	    Context jndiContext = getInitialContext(myReplica.getServerName(),
						    Integer.parseInt(myReplica.getServerPort()));
	    Object ref = jndiContext.lookup("AccountHomeRemote");
	    AccountHomeRemote home = (AccountHomeRemote)
		PortableRemoteObject.narrow(ref,AccountHomeRemote.class);
	    acctBean = home.create();
	    
	    // Connect to the Auction Bean
	    ref = jndiContext.lookup("AuctionHomeRemote");
	    AuctionHomeRemote AucHome = (AuctionHomeRemote)
		PortableRemoteObject.narrow(ref,AuctionHomeRemote.class);
	    aucBean = AucHome.create();
	    
	} catch(Exception e) {
	    alive = false;
        }

	return alive;
    }

    /* Get a new initial context that sets up the client to find
     * the JNDI server.
     *
     * @param jndiServer The URL of the JNDI server to connect to
     * @param port The default port for this connection
     * @return A handle to the context that JNDI should use
     */
    private Context getInitialContext(String jndiServer, int port) throws javax.naming.NamingException {
        Properties env = new Properties();
        
	// Set the port if no default is given
        if(jndiServer.indexOf(":") == -1)
            jndiServer = jndiServer + ":" + Integer.toString(port);
        
        env.setProperty(Context.INITIAL_CONTEXT_FACTORY,
                "org.jnp.interfaces.NamingContextFactory");
        env.setProperty(Context.URL_PKG_PREFIXES,
                "org.jboss.naming:org.jnp.interfaces");
        env.setProperty(Context.PROVIDER_URL, jndiServer);
        
        return new InitialContext(env);
    }    

    /* Main program globals */
    private boolean loop             = true;
    private boolean debug            = false;          /* Wether this server should print debug info */
    private RepCommonData myReplica  = null;           /* This server's replica data */
    private int cod                  = 1;              /* Cause of Death (initially unknown) */
    private AuctionRemote aucBean    = null;
    private AccountRemote acctBean   = null;

    private Context globalJndi       = null;
    private String jndiName          = null;
    
    /* Probe info for HP timing */
    private Time time                = null;
    private Probe probe1             = null;

    /* Cause of Death (COD) constant values */
    public static final int COD_UNKNOWN        =  1;   /* Unknown */
    public static final int COD_PCRASH         =  2;   /* Process Crash */
    public static final int COD_SCRASH         =  3;   /* Server/Node Crash */
    public static final int COD_EXCEPTION      =  4;   /* Server Exception thrown */
    public static final int COD_IMPRSTART      =  5;   /* Server did not properly start */

    // Default ports and boot messages
    private static final int LOCAL_JNDI_PORT    = 6729;
    private static final int HEARTBEAT_PERIOD   = 400;
    private static final String SERVER_EXCEPTION= "java.net.BindException";   /* Server exception string */
    private static final String JBOSS_STARTED   = "Started in";  /* Server started string */
}

