
package ejbay.replication;

import ejbay.account.AccountHomeRemote;
import ejbay.account.AccountRemote;
import ejbay.auction.*;
import ejbay.Probe.*;
import ejbay.util.*;

import javax.naming.InitialContext;
import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.Binding;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import java.io.*;
import java.util.Properties;
import java.lang.*;
import java.util.*;

import java.lang.reflect.Array;

public class ReplicationManager {

    /* Main method that starts and runs the replication manager */
    public static void main(String [] args) {
        BufferedReader cmdLine=new BufferedReader(new InputStreamReader(System.in));
        boolean running = true;
        FaultInjector fInj;
        int initCnt = 0;

        // Check to make sure that the necessary input arguments are provided
        if((args.length != 1 && args.length != 2 && args.length != 3) || (args[0].compareTo("") == 0)) {
            System.out.println("Error: Incorrect parameters passed");
            System.out.println(" usage: java ReplicationManager <JNDI server:port> <optional:server deploy dir>");
            System.exit(1);
        }
        
        // Conect to the global JNDI server that should have been passed in
        try
        {
            globalJndi = getInitialContext(args[0], GLOBAL_JNDI_PORT);
            
        }catch(javax.naming.NamingException ne){
            //Given global JNDI is no good, so just error and quit
            System.out.println("Error: Could not connect to the global JNDI server");
            System.out.println("       make sure the server is started and the\n" +
                               "       proper address was given");
            System.exit(1);
        }
        
        // Set any parameters that are to be passed to the server
        serverParams = args[1];

        String[] q  = new String[1];
        q[0] = "How many replicas would you like (max is " + serverList.availServers() + ")? ";
        String[] a = new String[1];

        // Ask the user for the number of replicas or grab it from the command line input
        if(args.length != 3)
            a = AskUser(q);
        else
            a[0] = args[2];

        // Begin probes
        time = new Time();
        probe1 = new Probe("repmngr_fault_corrected_time.txt");

        // Parse the number of replicas value entered to ensure it is valid
        try
        {
            numReplicas = (new Integer(a[0])).intValue();
            
            // No good, we can't support that many replicas
            if(numReplicas > serverList.availServers())
            {
                System.out.println("Number greater than maximum, only creating "
                                   + serverList.availServers() + " servers");
                numReplicas = serverList.availServers();
            }
            else if(numReplicas == -1)  // Start a server that prints debug output to the console
            {
                System.out.println("Starting one server in debug mode");
                numReplicas = 1;
                debug = true;
            }
            else if(numReplicas < -1 || numReplicas == 0)  // Need positive replicas
            {
                System.out.println("Error: You must create at least 1 replica");
                System.exit(1);
            }
        } catch(NumberFormatException nfe) {
            System.out.println("Error: " + a[0] + " is not a number");
            System.exit(1);
        }

        /* Initialize the primary to be undefined */
        try
        {
            globalJndi.rebind("primaryNum", "-1");
        }catch(javax.naming.NamingException ne)
        {
            System.out.println("JNDI Server not responding, exiting...");
            System.exit(2);
        }

        /* Start the initial set of servers sequentially, waiting for one to boot fully before going to the next */
        heartBeats = new ArrayList();
        while(initCnt < numReplicas)
        {
            HeartBeat curThread;
            if(!createReplica(initCnt))
            {
                System.out.println("Error: Cannot create replicas, check config and try again");
                System.exit(1);
            }
            curThread = waitLastReplica(); // Wait for replica to finish booting
            initCnt++;

            if(!curThread.isAlive())  // If it didn't boot properly, possibly ban and try again
            {
                RepCommonData temp = curThread.getReplica();
                int cod = curThread.getCauseOfDeath();
                    
                if(cod == HeartBeat.COD_SCRASH || cod == HeartBeat.COD_EXCEPTION)
                {
                    serverList.blockServer(temp);
                    if(serverList.availServers() < numReplicas)
                    {
                        System.out.println("Error: Not enough good servers to support the required number of replicas, forcing quit");
                        System.exit(1);
                    }
                }
                // Remove server from list and decrement counter so it is started next time
                curThread.destroy();
                killReplica(temp, false);
                heartBeats.remove(heartBeats.indexOf(curThread));
                
                initCnt--;
            }
        }
        try
        {
            // Set the primary to be the first server and reset the number of primarys count
            globalJndi.rebind("primaryNum", "0");
	    globalJndi.rebind("primaryCnt", new Integer(0));
        }catch(javax.naming.NamingException ne)
        {
            System.out.println("JNDI Server not responding, exiting...");
            running = false;
        }

        // Start the fault injector data entry console for the user
        fInj = new FaultInjector(heartBeats, args[0], time);
        if(running) fInj.start();

        // Monitor the heart beat threads and handle relaunching, etc.
        while(fInj.isAlive() && running)
        {
            for(int i = 0; i < numReplicas; i++)
            {
                HeartBeat curThread = (HeartBeat)heartBeats.get(i);
                if(!curThread.isAlive())
                {
                    // If a heartBeat failed, kill that replica and restart it
                    RepCommonData temp = curThread.getReplica();
                    int cod = curThread.getCauseOfDeath();
                    
                    // Ban this replica if it was a node crash or bind exception
                    if(cod == HeartBeat.COD_SCRASH || cod == HeartBeat.COD_EXCEPTION)
		    {
                        serverList.blockServer(temp);
			if(serverList.availServers() < numReplicas)
			{
			    System.out.println("Error: Not enough good servers to support the required number of replicas, forcing quit");
			    running = false;
			    break;
			}
                    }
                    // Kill off the dead server and then re-launch it
                    curThread.destroy();
                    int repNum = killReplica(temp, false);
                    heartBeats.remove(i);
                    createReplica(repNum);

                    // Update the primary as is neccessary
                    if(temp.isPrimary())
                    {
                        System.out.println("Primary server " + temp.getServerName() 
                                           + " failed, finding a new one");
                        updatePrimary();
                    }
                    probe1.record(String.valueOf(time.getElapsedTime()));
                }
            }
            
            // Sleep for a bit before checking the replicas again
            try {
                Thread.sleep(LOOP_DELAY);
            }catch(Exception e){}

        }

        // We're done so make sure all the threads are dead and clean things up
        if(fInj.isAlive())
            fInj.destroy();

        System.out.println("Quitting\nCleaning up JNDI");
	
        // Clean up the global JNDI
	try
	{
	    globalJndi.unbind("primaryNum");
	    globalJndi.unbind("primaryCnt");
	}catch(javax.naming.NamingException ne){}

        // Kill all threads
        int size = heartBeats.size();
        for(int i = size-1; i >= 0; i--)
        {
            HeartBeat tempThread = (HeartBeat)heartBeats.get(i);
            tempThread.destroy();
            try {
                tempThread.join(JOIN_DELAY);
            } catch(java.lang.InterruptedException ie){}
        }

        // Kill all the replicas
        killAllReplicas();
        probe1.output();
        System.out.println("Done.");
    }
    
    /** 
     * Prompts the standard input using a provided list of questions and
     * returns a list of the users answers for processing by the application
     *
     * @param questions A list of the questions to be asked in a string array
     * @return A list of the answers to the given questions as a string array
     */
    public static String[] AskUser(String[] questions) {
        String[] answers = new String[questions.length];
        BufferedReader input = new BufferedReader(new InputStreamReader(System.in));
        
        try {
            String inputString;
            for (int i = 0; i < questions.length; i++) {
                System.out.print(questions[i]);
                inputString = input.readLine();
		if (inputString.length() == 0)
                    inputString = " ";
                answers[i] = inputString;
	    }
        } catch (IOException ioe) {
            System.out.println("IO error trying to read your input!");
            System.exit(1);
        }
        return answers;
    }

    /** 
     * Get a new initial context that sets up the client to find
     * the JNDI server.
     *
     * @param jndiServer The URL of the JNDI server to connect to
     * @param port The default port number for this connection
     * @return A handle to the context that JNDI should use
     */
    public static Context getInitialContext(String jndiServer, int port) throws javax.naming.NamingException {
        Properties env = new Properties();
        
        // If no port number given use the passed in default
        if(jndiServer.indexOf(":") == -1)
            jndiServer = jndiServer + ":" + Integer.toString(port);
        
        env.setProperty(Context.INITIAL_CONTEXT_FACTORY,
                "org.jnp.interfaces.NamingContextFactory");
        env.setProperty(Context.URL_PKG_PREFIXES,
                "org.jboss.naming:org.jnp.interfaces");
        env.setProperty(Context.PROVIDER_URL, jndiServer);
        
        return new InitialContext(env);
    }
        
    /**
     * This function will look at the list of launched servers
     * and wait for the most recently finished one to finish
     * booting.
     *
     * @return A reference to the replica that was waited on
     */
    public static HeartBeat waitLastReplica()
    {
        int repNum = heartBeats.size()-1;
        boolean booted = false;
        HeartBeat tempThread = (HeartBeat)heartBeats.get(repNum);

        while(!booted && tempThread.isAlive())
        {
            //ask heartbeat thread if it's replica is booted
            booted = tempThread.isBooted();
            try{
                Thread.sleep(STARTUP_DELAY);
            }catch(InterruptedException ie){}
        }

        return tempThread;
    }

    /**
     * Go through the list of current replicas and find a new
     * replica to be the primary. If possible this will be
     * a replica that is already booted.
     */
    public static void updatePrimary()
    {
        int i;
        
        // Go through the list of replicas and find the first
        // booted replica
        for(i=0; i < numReplicas; i++)
        {
            RepCommonData temp = serverList.findServer(i);
            
            if(temp.isBooted())
            {
                serverList.setPrimary(i);
                break;
            }
        }
        if(i == numReplicas)  //No booted replica so just use the first replica
            serverList.setPrimary(0);

        // Now that we have a new primary, set the JNDI to agree
        updateGlobalJndi(serverList.getPrimaryIndex());
        System.out.println("New primary is " + (serverList.getPrimary()).getServerName());
    }

    /**
     * Updates the current primary value in the global JNDI
     *
     * @param primNum The number of the new primary to be stored
     *                in the global JNDI
     */
    public static void updateGlobalJndi(int primNum)
    {
        try
        {
            globalJndi.rebind("primaryNum", String.valueOf(primNum));
	    int primaryCnt = ((Integer)globalJndi.lookup("primaryCnt")).intValue();
	    primaryCnt++;
	    globalJndi.rebind("primaryCnt", new Integer(primaryCnt));
        }catch(javax.naming.NamingException ne){}    
    }
    
    /**
     * Creates a new replica with the given server number.  This will
     * handle launching the process, updating the global JNDI, and
     * spawning a heartBeat thread to monitor this replica.
     *
     * @param servNum The number of the server to be written to JNDI
     * @return True if this server is correctly started, false otherwise
     */
    public static boolean createReplica(int servNum)
    {
        String serv = serverList.findOpenServer();
        boolean isPrimary = false;
        String cmd;
        int serverNum = 0;
        RepCommonData repData = null;

        // If no server available, just exit
        if(serv == null)
        {
            System.out.println("Cannot create more replicas!");
            return false;
        }

        // Sleep for a small period to hopefully prevent creating
        // replicas concurrently
        try
        {
            Thread.sleep(CREATE_DELAY);
        }catch(InterruptedException ie){}
        
        // Make the first server the primary
        if(serverList.numServers() == 0)
            isPrimary = true;
        
        try
        {
            // SSH into the server and run our start server script
            cmd = SSH_PATH + serv + " " + START_CMD + " " + serverParams;
            //System.out.println("Trying: "+ cmd);
            Process p = Runtime.getRuntime().exec(cmd);
	    
            // Store this replica's data in a datastructure for future use
            repData = new RepCommonData(serv, Integer.toString(LOCAL_JNDI_PORT), p, isPrimary);            
            serverNum = serverList.addServer(repData, servNum);           
        }
        catch(IOException ioe)
        {
            System.out.println("Could not launch server on " + serv);
            return false;
        }
        
        // Reset the server's information in the global JNDI
        // this will be corrected later by the heartbeat thread
        try
        {           
            globalJndi.rebind("server" + serverNum, "null");
        } catch(Exception e) {
            killReplica(repData, false);
            return false;        
        }
        
        System.out.println("Launching server: " + repData.getServerName() + ":" + repData.getServerPort());

        //Spawn Heartbeat Thread        
        HeartBeat nThread = new HeartBeat(repData, globalJndi, "server" + serverNum, debug, time);
        heartBeats.add(nThread);
        nThread.start();

	return true;
    }    

    /**
     * Kills the replica with the given replication data and will either simply
     * reset the data in the global JNDI or remove it depending on the make
     * clean variable
     *
     * @param repData The replica to be killed
     * @param mkClean If true, this replica will be removed from the JNDI
     *                entirely. Otherwise, it will just be set to null
     * @return The number of the replica that was just killed
     */
    public static int killReplica(RepCommonData repData, boolean mkClean)
    {
        int replicaNum = -1;

        try
        {           
            // Run our kill script with the given server's name
            String cmd = KILL_CMD + " " + repData.getServerName();
            //System.out.println("Trying: "+ cmd);
            Process p = Runtime.getRuntime().exec(cmd);            
            
            p.waitFor();
        }catch(Exception e)
        {
            System.out.println("Could not kill server " + repData.getServerName());
            return -1;
        }

        // Find the server in the list and update the global JNDI
        try
        {
            int thisServer = serverList.findServerIndex(repData);
            
            if(thisServer == -1)
                return -1;
            
            if(!mkClean)
                globalJndi.rebind("server" + thisServer, "null");
            else
                globalJndi.unbind("server" + thisServer);

            replicaNum = serverList.removeServer(repData);  
            
        } catch(Exception e) {
            //Could not remove from global JNDI, it's probably crashed
            //so do we care??
        }        
        return replicaNum;
    }
    
    /**
     * Kill every replica currently known to be running and monitored
     * by the replication manager
     */
    public static void killAllReplicas()
    {
        int servers = serverList.numServers();

        System.out.println("Killing all replicas...");
        
        // Call killReplica on every server in the list
        for(int i=servers-1; i >=0; i--)
        {
            killReplica(serverList.findServer(i), true);
        }
    }
    
    // Global connection variables
    static Context globalJndi;    
    static String serverParams;
    static Time time;
    public static ServList serverList = new ServList();
    private static int numReplicas = 0;
    private static boolean debug = false;
    private static Probe probe1;
    private static ArrayList heartBeats         = null;

    // Various Remote system command setup stuff
    private static final String SSH_PATH        = "/usr/local/bin/rsh -f ";
    private static String CWD                   = System.getProperty("user.dir", "");
    private static String START_CMD             = CWD + "/runServerHelper";
    private static String KILL_CMD              = CWD + "/killAllGamesJava";

    // Various delays
    private static final int STARTUP_DELAY      = 1000;
    private static final int LOOP_DELAY         = 1000;
    private static final int CREATE_DELAY       = 200;
    private static final int JOIN_DELAY         = 3000;

    // JNDI Port Numbers
    private static final int LOCAL_JNDI_PORT    = 6729;
    private static final int GLOBAL_JNDI_PORT   = 59080;
}

