package ejbay.replication;

import java.io.*;
import java.lang.*;
import java.util.*;


public class ServList {
    public final int NUM_SERVERS    = 14;
    private String[] serverNames    = new String[NUM_SERVERS];
    private boolean[] serverBlocked = new boolean[NUM_SERVERS];
    private boolean[] serverBlckPrv = new boolean[NUM_SERVERS];
    private RepCommonData[] serverData = new RepCommonData[NUM_SERVERS];
    private int numSrvs             = 0;

    public ServList(){
        serverNames = new String[] {
            "settlers.lab.ece.cmu.local",
            "othello.lab.ece.cmu.local",
            "boggle.lab.ece.cmu.local",
            "drlucky.lab.ece.cmu.local",
            "arkhamhorror.lab.ece.cmu.local",
            "risk.lab.ece.cmu.local",
            "chess.lab.ece.cmu.local",
            "girltalk.lab.ece.cmu.local",
            "roborally.lab.ece.cmu.local",
            "go.lab.ece.cmu.local",
            "diplomacy.lab.ece.cmu.local",
            "devilbunny.lab.ece.cmu.local",
            "stratego.lab.ece.cmu.local",
            "clue.lab.ece.cmu.local"
        };
	
        for(int i = 0; i < NUM_SERVERS; i++)
        {
            serverBlocked[i] = false;
            serverBlckPrv[i] = true;
            serverData[i] = null;
        }
    }

    /**
     * Block the given replica from being used by the server list
     * when a call to find open server is issued
     *
     * @param repData The replica to be blocked
     */
    public void blockServer(RepCommonData repData)
    {
        String serverName = repData.getServerName();

        for(int i = 0; i < NUM_SERVERS; i++)
        {
            if(serverNames[i].equalsIgnoreCase(serverName))
                serverBlocked[i] = true;
        }
    }

    /* This function will purge all servers from the blocked list, essentially allowing
     * servers to be used again. It also checks to see if this block list is identical
     * to the list the last time a purge was run. If so it returns false to indicate that
     * these servers may really be dead.
     *
     * @return true if the blocked servers are the same as they were the last time purge
     *         was invoked, and false otherwise.
     */
    public boolean purgeBlocked()
    {
        boolean diff = false;

        for(int i = 0; i < NUM_SERVERS; i++)
        {
            if(serverBlocked[i] != serverBlckPrv[i])
                diff = true;
            serverBlckPrv[i] = serverBlocked[i];
            serverBlocked[i] = false;           
        }

        return diff;
    }

    /**
     * Get an object representing the current primary
     *
     * @return The replica object representing the primary
     */
    public RepCommonData getPrimary(){

        int index = getPrimaryIndex();
        
        if(index < 0)
            return null;

        return serverData[index];
    }

    /**
     * Get the index of the current primary according to the
     * server list
     *
     * @return The index of the current primary
     */
    public int getPrimaryIndex(){
        
	for(int i=0; i < numServers();i++){
	    if(serverData[i] != null && 
               serverData[i].isPrimary()) 
                return i;
	}
	return -1;
    }

    /**
     * Set the primary to be the server with the given index
     * and make sure to unset any current primarys
     *
     * @param index The index of the new primary
     */
    public void setPrimary(int index){
        
        if(index >= this.numServers() || index < 0)
            return;

        RepCommonData curPrim = this.getPrimary();
        RepCommonData temp;

        if(curPrim != null)
        {
            int primary = findServerIndex(curPrim);

            serverData[primary].setBackup();
        }

        if(serverData[index] != null)
            serverData[index].setPrimary();
    }

    /**
     * Set the replica with the given index to be booted
     *
     * @param index The index of the server to alter
     */
    public void setBooted(int index)
    {
        RepCommonData temp;

        if(index >= this.numServers() || index < 0)
            return;
        
        if(serverData[index] != null)
            serverData[index].setBooted();
    }

    /**
     * Find any server that is running currently according to the server
     * list.
     *
     * @return A string representing the name of the server
     */
    public String findOpenServer(){
	boolean found = false;
        Random r = new Random(System.currentTimeMillis());
        int serv = 0, cur = 0;
        
        if(this.numServers() >= this.availServers())
            return null;

        //serv = r.nextInt(this.availServers()-this.numServers());
        
	for(int i=0;i<NUM_SERVERS;i++){
	    String serverName = this.serverNames[i];
	    for(int j=0;j<numServers();j++){
		if(serverData[j] != null &&
                   serverName.equals((serverData[j]).getServerName()))
                    found=true;
	    }
	    if (!found)
            {
                if(!serverBlocked[i])
                {
                    if(cur == serv)
                        return serverName;
                    else
                        cur++;
                }
            }
            found = false;

	}
	return null;
    }

    /**
     * Locate a server that is currently running as recorded in the server
     * list.
     *
     * @return The replica that was chosen
     */
    public RepCommonData findUsedServer(){ 
        Random r = new Random(System.currentTimeMillis());
        int serv, cur = 0;

        if(this.numServers() == 0)
            return null;

        serv = r.nextInt(this.numServers());
        
	for(int i=0;i<NUM_SERVERS;i++){            
	    String serverName = this.serverNames[i];
	    for(int j=0;j<numServers();j++){
		if(serverData[j] != null && 
                   serverName.equals((serverData[j]).getServerName()))
                {
                    if(cur == serv)
                        return serverData[j];
                    else
                        cur++;
                }
	    }
	}
	return null;
    }

    /**
     * Adds a server with the given replica data to the list at the index
     * specified
     *
     * @param server The replica to be inserted
     * @param index The index to put this server at
     * @return The index that this server was inserted at (should be the same
     *         as the index input parameter)
     */
    public int addServer(RepCommonData server, int index){
	serverData[index] = server;
        numSrvs++;

	return index;
    }

    /**
     * Remove the server that matches the given replica data
     * from the server list and return its previous index
     *
     * @param server The replica to be removed
     * @return The location of where the server was in the list
     */ 
    public int removeServer(RepCommonData server){
	int index = findServerIndex(server);
        
        serverData[index] = null;
        numSrvs--;

	return index;
    }

    /**
     * Find the index of a server in the server list where
     * the given replica is located.
     *
     * @param repData The replica to find
     * @return The index of this replica
     */
    public int findServerIndex(RepCommonData repData)
    {
        for(int i=0; i < this.numServers(); i++)
        {
            if(repData.equals(findServer(i)))
                return i;
        }
        return -1;
    }
    
    /**
     * Finds a server in the server list based on the given
     * index
     *
     * @param index The index of the server to find
     * @return The replica located at this index
     */
    public RepCommonData findServer(int index){
        if(index >= numServers() || index < 0)
            return null;

	return serverData[index];
    }

    /**
     * Returns the number of currently running servers
     *
     * @return The number of running servers
     */
    public int numServers(){
	return numSrvs;
    }
    
    /**
     * Returns the number of available servers, taking into account
     * the total number of servers and removing the blocked servers.
     *
     * @return The number of available servers
     */
    public int availServers()
    {
        int avail = NUM_SERVERS;

        for(int i = 0; i < NUM_SERVERS; i++)
            if(serverBlocked[i])
                avail--;

        return avail;
    }
}
